//JetEngineCodeSnippets.cpp
//by Thunder Chicken
//August 14, 2022

//NOTE: THIS CODE WILL NOT COMPILE AS-IS! IT IS MEANT TO PROVIDE ACCESS TO CODE SNIPPETS
//AND GUIDANCE ON WHERE JET ENGINE CODE SHOULD BE LOCATED FOR DEVELOPERS TO IMPLEMENT IN 
//THEIR OWN ADD-ON CODE. PLEASE READ THE ASSOCIATED QUICK START GUIDE FOR MORE INFORMATION.

//Enjoy, and good luck with your add-on! If you have questions or problems, PM me through 
//Orbiter Forum (https://www.orbiter-forum.com). -TC

#define STRICT
#define ORBITER_MODULE

#include <orbitersdk.h>
#include <cstdio>
#include <cmath>

//Add-on developers must add JetEngineLibrary.h and JetEngineLibrary.lib to their local add-on
//directory and add these files to their solution.

#include "JetEngineLibrary.h"

// Calculate maximum air volume flow rate from inlet area and maximum air mass flow rate
// specified for engine. This requires iteratively solving once for air density and velocity
// at the inlet assuming stationary full throttle operation at sea level

const double air_volume_flow_max = JetEngineLibrary::JetEngine::GetJetEngineMaxAirFlow(inlet_diameter, air_mass_flow_max);

class Vessel: public VESSEL3
{
public:
	
	//Read afterburner status from scenario file
	void clbkLoadStateEx(FILEHANDLE scn, void *status);
	
	//Write afterburner status to scenario file
	void clbkSaveState(FILEHANDLE scn);
	
	//Define thrusters needed to implement jet engine model
	void clbkSetClassCaps(FILEHANDLE cfg);
	
	//Calculate thrust and Isp at each time step of simulation
	void clbkPreStep(double simt, double simdt, double mjd);
	
	//Implement key toggle of afterburner
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);

	//Add afterburner status flag to standard Orbiter HUD
	bool clbkDrawHUD(int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	
private:

	PROPELLANT_HANDLE main_fuel_tank;

	//Thrusters and thruster groups for jet engine implementation
	
	THRUSTER_HANDLE th_main;
	THRUSTER_HANDLE th_ab;
	THRUSTER_HANDLE th_dummy;

	THGROUP_HANDLE thg_main;
	THGROUP_HANDLE thg_ab;
	THGROUP_HANDLE thg_dummy;

	//Afterburner status boolean and int equivalent to enable writing afterburner status to scenario file

	bool afterburner = false;
	UINT afterburner_status = 0;
};

void Vessel::clbkLoadStateEx(FILEHANDLE scn, void *status) 
{ 
	char *line; 
	while (oapiReadScenario_nextline(scn, line)) 
	{ 
		//Read afterburner status from scenario file

		if (!_strnicmp(line, "afterburner_status", 18))
		{
			sscanf(line + 18, "%li", &afterburner_status);

			if (afterburner_status == 1)
			{
				afterburner = true;
			}
			else
			{
				afterburner = false;
			}

		}
		
	} 
} 

void Vessel::clbkSaveState(FILEHANDLE scn) 
{ 
	VESSEL3::clbkSaveState(scn);

	//Write afterburner status to scenario file

	if (afterburner == true)
	{
		afterburner_status = 1;
	}
	else
	{
		afterburner_status = 0;
	}
		
	oapiWriteScenario_int(scn, "afterburner_status", afterburner_status);

}

void Vessel::clbkSetClassCaps(FILEHANDLE cfg)
{
	//Make main propellant tank

	main_fuel_tank = CreatePropellantResource(main_fuel_tank_max);

	//Make main thruster
	//This is a user-defined thruster that is meant to set the thrust based on the output
	//of the jet engine model, which is a function of not only throttle position but of
	//ambient atmospheric conditions and Mach number. Variable max_thrust is user-defined 
	//in kN, so must be converted to N
	
	th_main = CreateThruster(_V(0, 0, -4.3), _V(0, 0, 1), max_thrust*1000, main_fuel_tank, Isp_max);
	thg_main = CreateThrusterGroup(&th_main, 1,THGROUP_USER);

	//Make afterburner thruster

	th_ab = CreateThruster(_V(0, 0, -4.3), _V(0, 0, 1), (max_thrust_ab - max_thrust) * 1000, main_fuel_tank, (Isp_max_ab - Isp_max));
	thg_ab = CreateThrusterGroup(&th_ab, 1, THGROUP_USER);

	//Make dummy thruster to take user throttle input to be
	//passed to main engine through the jet engine thrust model calculation
	//Note that this is linked to THGROUP_MAIN to get main engine user throttle input
	//As it is connected to the fuel tank, it will also allow proper indication of fuel level

	th_dummy = CreateThruster(_V(0, 0, -4.3), _V(0, 0, 1), 1.0, main_fuel_tank, INFINITY); //produces no thrust, consumes no fuel
	thg_dummy = CreateThrusterGroup(&th_dummy, 1, THGROUP_MAIN);

}

void Vessel::clbkPreStep(double simt, double simdt, double mjd)
{
	//----JET ENGINE IMPLEMENTATION

	//Get throttle input from user from dummy thruster

	double throttle = GetThrusterGroupLevel(thg_dummy);

	if (throttle < 1.0)
	{
		afterburner = false; //shuts off afterburner when not at military power
	}

	// Get local flight conditions from Orbiter

	double M_0 = GetMachNumber();	        //flight Mach number [-]
	double T_0 = GetAtmTemperature();       //ambient static temperature [K]
	double P_0 = GetAtmPressure();          //ambient static pressure [Pa absolute]
	double rho = GetAtmDensity();         	//ambient atmospheric density [kg/cubic meter]

	//Use inputs and flight conditions to calculate jet engine thrust

	double thrust = JetEngineLibrary::JetEngine::GetJetEngineThrust(throttle, afterburner, air_volume_flow_max, M_0, T_0, P_0, rho);

	//Calculate Isp of main engine and afterburner

	double Isp_main = JetEngineLibrary::JetEngine::GetJetEngineIsp(throttle, afterburner, thrust);
	double Isp_ab   = JetEngineLibrary::JetEngine::GetJetEngineAfterburnerIsp(throttle, afterburner, thrust);

	//Update Isp for user-defined main engine and afterburner

	SetThrusterIsp(th_main, Isp_main);
	SetThrusterIsp(th_ab, Isp_ab);

	//Update thruster level for user-defined main engine and afterburner

	SetThrusterGroupLevel(thg_main, fmin(thrust, max_thrust*1000) /(max_thrust*1000));
	SetThrusterGroupLevel(thg_ab, (thrust - (max_thrust*1000)) / ((max_thrust_ab - max_thrust) * 1000));

	//----END JET ENGINE IMPLEMENTATION

}

int Vessel::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
	if (!down)
	{
		return 0;
	}

	switch (key)
	{

	case OAPI_KEY_A: // Afterburner

		//afterburner flag is set to false when throttle is reduced to < 1 in clbkPreStep

		if (afterburner == false && GetThrusterGroupLevel(thg_dummy) == 1.0)
		{
			afterburner = true; //turns on afterburner when at military power
		}
		else if (afterburner == true && GetThrusterGroupLevel(thg_dummy) == 1.0)
		{
			afterburner = false; //shuts off afterburner when at military power
		}
		return 1;

	return 0;
	}
}

bool Vessel::clbkDrawHUD(int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// Draw the default HUD

	VESSEL3::clbkDrawHUD(mode, hps, skp);

	// Add text to indicate afterburner status

	int x_coord = 27; // x coordinate from top left
	int y_coord = 155; // y coordinate from top left

	skp->SetOrigin(0, 0);

	if (afterburner == true)
	{
		skp->SetTextColor(0x0066FF66); // 0x00BBGGRR
		char abuf[256];
		sprintf(abuf, "AFTERBURNER ON");
		skp->Text(x_coord, y_coord, abuf, strlen(abuf));
	}
	else
	{
		skp->SetTextColor(0x0066FF66); // 0x00BBGGRR
		char abuf[256];
		sprintf(abuf, "AFTERBURNER OFF");
		skp->Text(x_coord, y_coord, abuf, strlen(abuf));
	}

	return true;
}